#include "stdafx.h"

#include <stdio.h>
#include <time.h>

//-- const int PAINTTIMERID = 537;
//-- todo create a timer message which ensures painting occurs occasionally

//-------------------------------------------------------------------------------
// dispatch NI callbacks to correct session 
//........................
int __cdecl GenDispatch( int iSamples, void * pData, void * pDigData, void * pSession, int iSetPos ) { 
	if( pSession == NULL )
		return -1;
	if( iSamples < 0 ) {	// less data used than provided, 'rewind' the counter.
		((CSession *)pSession)->m_iAnaGenPos += iSamples;
		((CSession *)pSession)->m_iDigGenPos += iSamples;
		return -2;
		}
	((CSession *)pSession)->GenData( iSamples, pData, pDigData, iSetPos );
	return 0;
	}

//........................
int __cdecl AcqDispatch( int iSamples, void * pData, void * pDigData, void * pSession ) { 
	if( pSession != NULL ) 
		((CSession *)pSession)->AcqData( iSamples, pData, pDigData );
	return 0;
	}

//........................
int __cdecl TermAcqDispatch( void * pSession ) { 
	if( pSession != NULL ) 
		((CSession *)pSession)->AcqCompleted();
	return 0;
	}

int __cdecl TermGenDispatch( void * pSession ) { 
	if( pSession != NULL ) 
		((CSession *)pSession)->GenCompleted();
	return 0;
	}

//-------------------------------------------------------------------------------
CDigoutEvent :: CDigoutEvent( int iAcqPos, int iLine, int iStatus ) { 
	m_iAcqPos = iAcqPos;
	m_iLine = iLine; 
	m_iStatus = iStatus;
	}

//-------------------------------------------------------------------------------
CSessionPanel :: CSessionPanel( int iWid, int iLft, int iTop ) { 
	m_hBitmap = NULL;
	m_hdcBitmap = NULL;
	m_iWid=iWid;
	m_iLft=iLft;
	m_iTop=iTop;
	m_bIconized=false;
	}
	
CSessionPanel :: ~CSessionPanel() { 
	DeleteBitmap();
	}
	
void CSessionPanel ::SetupBitmap( HDC hdc ) {				//-- TODO - call this if the window is resized 
	//-- setup display panel buffer
	DeleteBitmap();
	m_hdcBitmap = CreateCompatibleDC( hdc );
	m_hBitmap = CreateCompatibleBitmap( hdc, m_iWid, m_iHgt );

	SelectObject( m_hdcBitmap, m_hBitmap );
	PatBlt( m_hdcBitmap, 0, 0, m_iWid, m_iHgt, WHITENESS );
	m_iBitmapOffset = 0;
	}

	
void CSessionPanel :: DeleteBitmap() { 
	if( m_hBitmap != NULL ) {
		DeleteObject( m_hBitmap );
		DeleteDC( m_hdcBitmap );
		m_hBitmap = NULL;
		}
	}
	
int CSessionPanel :: YPos(double dValue) {	// position on screen of dValue based on m_dMin and m_dMax
	return min( m_iHgt-2, max( 1, m_iHgt-int((dValue-m_dMinDisp)/DispRange()*m_iHgt) ) );
	//return min( m_iBtm, max( m_iTop, m_iBtm - int((dValue-m_dMinDisp)/DispRange()*max(1,m_iBtm-m_iTop) ) ));
	}

bool CSessionPanel :: LoadFromFile( FILE * h ){ 
	int iVer;
	if( fread( &iVer, sizeof(m_iHgt), 1, h ) != 1 
		|| fread( &m_dMinDisp, sizeof(m_dMinDisp), 1, h ) != 1 
		|| fread( &m_dMaxDisp, sizeof(m_dMaxDisp), 1, h ) != 1 
		|| fread( &m_iPage, sizeof(m_iPage), 1, h ) != 1 
		|| fread( &m_iChannelType, sizeof(m_iChannelType), 1, h ) != 1 
		|| fread( &m_iChannelIndex, sizeof(m_iChannelIndex), 1, h ) != 1 
		|| fread( &m_iLft, sizeof(m_iLft), 1, h ) != 1 
		|| fread( &m_iTop, sizeof(m_iTop), 1, h ) != 1 
		|| fread( &m_iWid, sizeof(m_iWid), 1, h ) != 1 
		|| fread( &m_iHgt, sizeof(m_iHgt), 1, h ) != 1 
		) 
		return false;
	return true;
	}

bool CSessionPanel :: SaveToFile( FILE * h ){ 
	int iVer=1;
	return fwrite( &iVer, sizeof(iVer), 1, h ) == 1 
		&& fwrite( &m_dMinDisp, sizeof(m_dMinDisp), 1, h ) == 1 
		&& fwrite( &m_dMaxDisp, sizeof(m_dMaxDisp), 1, h ) == 1 
		&& fwrite( &m_iPage, sizeof(m_iPage), 1, h ) == 1 
		&& fwrite( &m_iChannelType, sizeof(m_iChannelType), 1, h ) == 1 
		&& fwrite( &m_iChannelIndex, sizeof(m_iChannelIndex), 1, h ) == 1 
		&& fwrite( &m_iLft, sizeof(m_iLft), 1, h ) == 1 
		&& fwrite( &m_iTop, sizeof(m_iTop), 1, h ) == 1 
		&& fwrite( &m_iWid, sizeof(m_iWid), 1, h ) == 1 
		&& fwrite( &m_iHgt, sizeof(m_iHgt), 1, h ) == 1 
		;
	}

void CSessionPanel :: SetupAnaAcqPanel( const char * pchName, const char * pchUnits, int iHgt, double dMin, double dMax, double dMinDisp, double dMaxDisp, 
							bool bVisible, int iChnlIndex, int iPage ) { 
	m_cName.Set( pchName );
	m_cUnits.Set( pchUnits );
	m_iHgt = iHgt;
	m_dMin = dMin;
	m_dMax = dMax;
	m_dMaxDisp = dMaxDisp;
	m_dMinDisp = dMinDisp;
	m_iPage = iPage;
	m_iChannelType = CT_AI;
	m_iChannelIndex = iChnlIndex;
	}

void CSessionPanel :: SetupAnaGenPanel( const char * pchName, const char * pchUnits, bool bVisible, int iChnlIndex, int iPage ) { 
	m_cName.Set( pchName );
	m_cUnits.Set( pchUnits );
	m_iHgt = ANAGENPANEL_HEIGHT;
	m_dMin = -1.0;
	m_dMax = 1.0;
	m_dMinDisp = -1.0;
	m_dMaxDisp = 1.0;
	m_iPage = iPage;
	m_iChannelType = CT_AO;
	m_iChannelIndex = iChnlIndex;
	}

void CSessionPanel :: SetupDigAcqPanel( BYTE * abLines, int iPage ) { 
	m_cName.Set( "DigAcq" );
	m_cUnits.Set( "On/Off" );
	m_iHgt = DIGACQPANEL_HEIGHT;
	m_dMin = -1.0;
	m_dMax = 1.0;
	m_dMinDisp = -1.0;
	m_dMaxDisp = 1.0;
	m_iPage = iPage;
	m_iChannelType = CT_DI;
	m_iChannelIndex = DIGGENCHANNELID;

	m_iMaxLineAcq=0;
	for( int iLine=0; iLine<MAXDIGLINES; ++iLine ) 
		if( abLines[iLine]==DOMODE_INPUT ) 
			m_abLineSelectAcq[m_iMaxLineAcq=iLine]=true;
		else 
			m_abLineSelectAcq[iLine]=false;
	}
			
void CSessionPanel :: SetupDigGenPanel( BYTE * abLines, int iPage ) {
	m_cName.Set( "DigGen" );
	m_cUnits.Set( "On/Off" );
	m_iHgt = DIGGENPANEL_HEIGHT;
	m_dMin = -1.0;
	m_dMax = 1.0;
	m_dMinDisp = -1.0;
	m_dMaxDisp = 1.0;
	m_iPage = iPage;
	m_iChannelType = CT_DO;
	m_iChannelIndex = -1;

	m_iMaxLineGen=0;
	for( int iLine=0; iLine<MAXDIGLINES; ++iLine ) 
		if( abLines[iLine]==DOMODE_PROTOCOL ) 
			m_abLineSelectGen[m_iMaxLineGen=iLine]=true;
		else
			m_abLineSelectGen[iLine]=false; 
	}		
		
double CSessionPanel :: DispRange() { 
	if ( m_dScaling > 0 )
		return max( 0.00001, m_dMaxDisp-m_dMinDisp );
	else
		return min( -0.00001, m_dMaxDisp-m_dMinDisp );
	}
	
//-- Process message to update data display for new acquired data 
void CSessionPanel :: AIUpdate( int iSampPerPixel, double * adQueue, 
			int iChans, int iChan, int iQueSize, int iQuePos, bool bRecord, HPEN penDraw,
			int iScanMode, int iDrawMode ) { 

	//-- calculate column mean
	double dSum=0.0;
	for( int iSamp=0; iSamp<iSampPerPixel; ++iSamp )
		dSum += adQueue[((iQuePos+iSamp)%iQueSize)*iChans+iChan];
	double dMean = dSum / iSampPerPixel;

	//-- clear column with pen set by caller ( White or time separator color ) 
	HPEN penWas = (HPEN) SelectObject( m_hdcBitmap, penDraw );
	MoveToEx( m_hdcBitmap, m_iBitmapOffset, 1, NULL );
	LineTo( m_hdcBitmap, m_iBitmapOffset, m_iHgt-1 );
	SelectObject( m_hdcBitmap, penWas );

	//-- short red indicator when bitmap 'Wraps'
	#ifdef _DEBUG
	if( 0==m_iBitmapOffset  ) { 
		SetPixel( m_hdcBitmap, 0, 2, RGB(255,0,0));
		SetPixel( m_hdcBitmap, 0, 3, RGB(255,0,0));
		SetPixel( m_hdcBitmap, 0, 4, RGB(255,0,0));
		}
	#endif

	//-- Line at top and bottom of panel ( red at top when recording ) 
	if( bRecord ) 
		SetPixel( m_hdcBitmap, m_iBitmapOffset, 0, RGB(128,0,0) );
		
	//-- clear data about to be overwritten (wrap)
	if( iScanMode == 2 ) { 
		int iCol = (m_iBitmapOffset+ERASEWIDTH)%m_iWid;
		HPEN penWhite = CreatePen( 0, 0, RGB(255,255,255) );
		penWas = (HPEN) SelectObject( m_hdcBitmap, penWhite );	
		MoveToEx( m_hdcBitmap, iCol, 0, NULL );
		LineTo( m_hdcBitmap, iCol, m_iHgt);
		SelectObject( m_hdcBitmap, penWas );
		DeleteObject( penWhite );
		}
	//-- clear underneath mean-value text
	if ( iScanMode != 1 ) { // incompatible with scrolling
		RECT rctMean;
		SetRect( &rctMean, m_iWid - MEANTEXT_BOXWID, 0, m_iWid, MEANTEXT_BOXHGT ); 
		FillRect( m_hdcBitmap, &rctMean, (HBRUSH) GetStockObject( BLACK_BRUSH ) );
		}

	//-- mean-value text
	if ( iScanMode != 1 ) { // incompatible with scrolling
		HFONT fntText = CreateFont(-MEANTEXT_FONTSIZE, //MulDiv(m_iSize,-100,76), 
							0 /*Width*/,
							0 /*escapement*/,
							0 /*nOrientation*/,
							FW_BOLD, //  : FW_REGULAR,
							FALSE /* Italic*/, 
							FALSE /* underline */,
							FALSE /* Strikeout */, 
							0,0,0,0, 
							(FF_MODERN | FIXED_PITCH), // : (FF_ROMAN | VARIABLE_PITCH)), 
							"Modern"); // : "Times New Roman"));
		HFONT fntWas = (HFONT) SelectObject( m_hdcBitmap, fntText );
		int iAlignWas = SetTextAlign( m_hdcBitmap, TA_BASELINE | TA_LEFT );
		COLORREF crWas = SetTextColor( m_hdcBitmap, MEANTEXT_COLOR );
		COLORREF bkWas = SetBkColor( m_hdcBitmap, RGB(0,0,0) );
		char pchText[40];
		sprintf( pchText, "%.6g", dMean / m_dScaling );
		TextOut( m_hdcBitmap, m_iWid - MEANTEXT_WID - ((pchText[0]=='-') ? (MEANTEXT_FONTSIZE/2) : 0), MEANTEXT_BASELINE, pchText, strlen( pchText ) );
		SetTextColor( m_hdcBitmap, crWas );
		SetBkColor( m_hdcBitmap, bkWas );
		SetTextAlign( m_hdcBitmap, iAlignWas );
		SelectObject( m_hdcBitmap, fntWas );
		DeleteObject( fntText );
		}
		
	//-- Data trace
	if ( ((m_dScaling >= 0.0) && (dMean>=m_dMax || dMean<=m_dMin))
	  || ((m_dScaling < 0.0) && (dMean<=m_dMax || dMean>=m_dMin)) ) {
		HPEN penSat = CreatePen( 0, 0, crSaturatedColor );
		penWas = (HPEN) SelectObject( m_hdcBitmap, penSat );	
		MoveToEx( m_hdcBitmap, m_iBitmapOffset, 1, NULL );
		LineTo( m_hdcBitmap, m_iBitmapOffset, m_iHgt-1);
		SelectObject( m_hdcBitmap, penWas );
		DeleteObject( penSat );
		}

	if( iDrawMode==3 ) {						//--- draw every point
		for( int iSamp=0; iSamp<iSampPerPixel; ++iSamp ) 
			SetPixel( m_hdcBitmap, m_iBitmapOffset, YPos( adQueue[((iQuePos+iSamp)%iQueSize)*iChans+iChan] ), crTraceColor );
		}
	else if( iDrawMode==2 ) {					//--- draw average
		SetPixel( m_hdcBitmap, m_iBitmapOffset, YPos(dMean), crTraceColor );
		}
	else										//--- (1) draw single point
		SetPixel( m_hdcBitmap, m_iBitmapOffset, YPos(adQueue[iQuePos*iChans+iChan]), crTraceColor );
		
	m_iBitmapOffset=(m_iBitmapOffset+1) % m_iWid;
	}							
			

void CSessionPanel :: DIUpdate( int iSampPerPixel, int * aiDigQueue, 
			int iQueSize, int iQuePos, int iLines, BYTE * aiDigAcqMap, 
			int iScanMode, int iDrawMode ) { 

	//-- clear data about to be overwritten (wrap)
	if( iScanMode == 2 ) { 
		int iCol = ( m_iBitmapOffset+ERASEWIDTH ) % m_iWid;
		HPEN penWhite = CreatePen( 0, 0, RGB(255,255,255) );
		HPEN penWas = (HPEN) SelectObject( m_hdcBitmap, penWhite);
		MoveToEx( m_hdcBitmap, iCol, 0, NULL );
		LineTo( m_hdcBitmap, iCol, m_iHgt);
		SelectObject( m_hdcBitmap, penWas );
		DeleteObject( penWhite );
		}
		
	int iPixPerLine = iLines ? (m_iHgt/iLines) : 0;
	COLORREF crBit;

	for( int iLine=0; iLine<iLines; ++iLine ) { 
		//--- set pixel color
		BYTE btSet=0;
		if( iDrawMode==1 )		
			crBit = aiDigQueue[iQuePos] & (1<<aiDigAcqMap[iLine]) ? crDigitalOn : crDigitalOff;
		else {						
			for( int iSamp=0; iSamp<iSampPerPixel && btSet!=3 ; ++iSamp )
				btSet |= aiDigQueue[(iQuePos+iSamp)%iQueSize] & (1<<aiDigAcqMap[iLine]) ? 2 : 1; 
			
			crBit = btSet==1 ? crDigitalOff : ( btSet==2 ? crDigitalOn : crDigitalTrn );
			}
			
		//---- fill in pixels.  If >2 pixels/DigLine, leave blank pixels between DigLines
		int iOffset=iPixPerLine*iLine;
		for( int iPixel=iPixPerLine>2?1:0; iPixel<iPixPerLine; ++iPixel ) 
			SetPixel( m_hdcBitmap, m_iBitmapOffset, m_iHgt-(iOffset+iPixel), crBit );
		}
	m_iBitmapOffset=(m_iBitmapOffset+1)%m_iWid;
	}

//-- Change the height of a panel
void CSessionPanel :: AdjustPanel( int iAdjust, int iX, int iY ) { 
	int iBtm = m_iTop+m_iHgt;
	int iRgt = m_iLft+m_iWid;

	if( iAdjust==CAP_TOP ) { 
		m_iTop=min( iY, iBtm-22 );
		m_iHgt=iBtm-m_iTop;
		}
	else if( iAdjust==CAP_BTM ) { 
		m_iHgt=max( iY-m_iTop, 22 );
		}
	else if( iAdjust==CAP_LFT ) { 
		m_iLft=min( iX+SCOPE_XPOS, iRgt-200 );
		m_iWid=iRgt-m_iLft;
		}
	else if( iAdjust==CAP_RGT ) { 
		m_iWid=max( iX-m_iLft, 200 );
		}
	}

//-- Change the height of a panel
void CSessionPanel :: MovePanel( int iX, int iY, RECT rctPic ) { 
	m_iTop=max(rctPic.top,min(rctPic.bottom-m_iHgt,iY));
	m_iLft=max(rctPic.left+SCOPE_XPOS,min(rctPic.right-m_iWid,iX));
	}

void CSessionPanel :: PlacePanel( int iLeft, int iTop, int iWid, int iHgt )
{
	m_iLft = iLeft;
	m_iTop = iTop;
	m_iWid = iWid;
	m_iHgt = iHgt;
}

void CSessionPanel :: DrawExtra( HDC hdc, CConfiguration * poCfg ) { 
	int iLeft = m_iLft-SCOPE_XPOS;
	Rectangle( hdc, iLeft-1, m_iTop, m_iLft+1+(m_bIconized?0:m_iWid), m_iTop+m_iHgt+1 );

	if( ( m_iChannelType==CT_AI || m_iChannelType==CT_DI ) && m_iHgt>=12 ) { // Minimize icon
		RECT rctMinimize;
		SetRect( &rctMinimize, m_iLft-SCOPE_XPOS, m_iTop+1, m_iLft-SCOPE_XPOS+ICONIZE_BOX_SIZE, m_iTop+ICONIZE_BOX_SIZE ); 
		FillRect( hdc, &rctMinimize, (HBRUSH) GetStockObject( BLACK_BRUSH ) );
		}

	if( m_iChannelType==CT_AI && m_iHgt>32 ) {	//-- add text range labels to AI panels 
		//-- Create resources
		HPEN penBlack = CreatePen( 0, 0, RGB(0, 0, 0) );
		HPEN penWas = (HPEN) SelectObject( hdc, penBlack );
		COLORREF crTextColorWas = SetTextColor( hdc, RGB(160,160,160) );
		COLORREF crBkWas = SetBkColor( hdc, RGB(255,255,255) );	// for text labels

		//-- draw text 
		char buff[100];
		double dScaling = poCfg->m_aoInChnCfg[m_iChannelIndex].m_dScaling;
		
		sprintf( buff, "%f", m_dMaxDisp * 2.0 / dScaling );
		ZTrim(buff);
		TextOut( hdc, iLeft+TITLE_WIDTH, m_iTop+1, buff, min(6,strlen(buff)) );
		
		sprintf( buff, "%f", m_dMinDisp * 2.0 / dScaling );
		ZTrim(buff);
		TextOut( hdc, iLeft+TITLE_WIDTH, m_iTop+m_iHgt-17, buff, min(6,strlen(buff)) );

		//-- cleanup 		
		SetBkColor( hdc, crBkWas );		
		SetTextColor( hdc, crTextColorWas );
		SelectObject( hdc, penWas );
		DeleteObject( penBlack );
		}
	}

//-------------------------------------------------------------------------------
CSession :: CSession() { 
	m_poCfg = NULL; 
	m_poAnaAcqTask = NULL;
	m_poAnaGenTask = NULL;
	m_poDigAcqTask = NULL;
	m_poDigGenTask = NULL;
	m_poCard = NULL;
	m_adQueue = NULL;
	m_aiDigQueue = NULL;
	m_oQDF.SetDataType(QTP_SHORT);
	m_bRecording=false;
	m_iGenChannel=0;
	m_bRecordDigout=false;
	m_iAcqPos=0;	
	m_iAnaGenPos=m_iDigGenPos=0;
	m_pFeedback=NULL;
	m_pAcqInit=NULL;
	m_pGenInit=NULL;
	m_bFeedback=false;
	m_bGainTable=false;
	m_pGainTable=NULL;
	m_pFeedback=NULL;
	m_aiDigAcqMap=NULL;
	m_aiDigGenMap=NULL;
	m_iPage=0;
	m_iPages=1;
	m_bQuick=false;
	}
	
CSession :: ~CSession() { 
	if( m_aiDigAcqMap != NULL ) 
		delete [] m_aiDigAcqMap;
		
	if( m_pFeedback != NULL )
		delete m_pFeedback;
		
	if( m_oQDF.Opened() ) {
		m_oQDF.WriteDataNotes();
		m_oQDF.Close();
		}
		
	if( m_pGainTable != NULL )
		delete m_pGainTable;
		
	}

void CSession :: SetupBitmaps( HDC hdc ) { 
	int iPanels = m_aPanels.Count();
	for( int iPanel=0; iPanel<iPanels; ++iPanel ) { 
		CSessionPanel * pPnl = (CSessionPanel *)(m_aPanels[iPanel]);
		if( pPnl->m_iChannelType==CT_AI || pPnl->m_iChannelType==CT_DI ) 
			pPnl->SetupBitmap( hdc );
		}
	}

bool CSession :: IsDigoutControl( int iLine ) { 
	return iLine>=0 
		&& iLine<CONFIG_DOROWS 
		&& ( m_poCfg->m_aiDigMode[iLine]==DOMODE_AUTO 
			|| m_poCfg->m_aiDigMode[iLine]==DOMODE_MANUAL
			|| m_poCfg->m_aiDigMode[iLine]==DOMODE_PROTOCOL );
	}

bool CSession :: IsDigoutControl( ) { 
	for( int iLine=0; iLine<CONFIG_DOROWS; ++iLine ) 
		if( IsDigoutControl(iLine) ) 
			return true;
	return false;
	}

bool CSession :: NewSession( CQUBIOSettings& oSet ) {
	m_cConfigName.Set( oSet.m_cConfigName.Buffer() );
	
	//-- Set default Gen parameters
	m_iGenBufferSize = oSet.m_iBufferSize;		// size in msec of each generate buffer
	m_iGenRate = oSet.m_iIORate;				// rate (in hz) of generation
	for( int i=0; i<SESSION_MAXGENCHN; ++i ) {
		m_acGenFile[i].Set("");
		m_abGenRepeat[i]=true;
		}
	m_cDigGenFile.Set("");
	m_bDigGenRepeat=true;
		
	//---------	Load parameters from registry
	char buf[_MAX_PATH+1], bufName[16];
	buf[_MAX_PATH]=0;
	DWORD dwSize;
	HKEY h;
	RegCreateKeyEx( HKEY_CURRENT_USER, "Software\\QubIO", 0, NULL, 0, KEY_ALL_ACCESS, NULL, &h, NULL );

	for( int iCh=0; iCh<4; ++iCh ) { 
		sprintf( bufName, "GenFile%d", iCh );
		if( ERROR_SUCCESS == RegQueryValueEx( h, bufName, 0, NULL, (LPBYTE)buf, &(dwSize=_MAX_PATH) ) ) {
			m_abGenRepeat[iCh]= ( buf[0]=='Y' );
			m_acGenFile[iCh].Set( buf+1 );
			}
		}
		
	if( ERROR_SUCCESS == RegQueryValueEx( h, "DigGenFile", 0, NULL, (LPBYTE)buf, &(dwSize=_MAX_PATH) ) ) {
		m_bDigGenRepeat = ( buf[0]=='Y' );
		m_cDigGenFile.Set( buf+1 );
		}
	RegCloseKey( h );

	//-- Set default Acq parameters
	m_iAcqBufferSize = oSet.m_iBufferSize;		// size in msec of each acquisition buffer
	m_iAcqRate = oSet.m_iIORate;				// rate (in hz) of acquisition
	m_iAcqSampPerPixel = 10;
	
	m_bGainTable=oSet.m_bGainTable;		
	m_cGainTable=oSet.m_cGainTable;
	m_iGainTableSource=oSet.m_iGainTableSource;
	m_iGainTableApply=oSet.m_iGainTableApply;
	
	m_bFeedback=oSet.m_bDefaultFeedback;
	m_cFeedback=oSet.m_cFeedbackName;
	m_iFeedbackSource=oSet.m_iFeedbackSource;
	m_iFeedbackApply=oSet.m_iFeedbackApply;

	m_bRecordDigout=oSet.m_bRecordDigout;	
	
	SYSTEMTIME st;
	GetLocalTime( &st );
	int(st.wYear); 
	int(st.wMonth);
	int(st.wDay) ;
	int(st.wHour);
	int(st.wMinute); 
	int(st.wSecond);

	struct tm tmNow;
    tmNow.tm_sec	= st.wSecond;
    tmNow.tm_min	= st.wMinute; 
    tmNow.tm_hour	= st.wHour;
    tmNow.tm_mday	= st.wDay ;
    tmNow.tm_mon	= st.wMonth-1;
    tmNow.tm_year	= st.wYear-1900; 

	char buffer[_MAX_PATH];
	if( strftime( buffer, _MAX_PATH, oSet.m_cAcqFileName.Buffer(), &tmNow ) > 0 ) 
		m_cAcqFile.Set( buffer );
	else 
		m_cAcqFile.Set("Data.qdf");

	m_aPanels.DeleteAll();
	return true;
	}

bool CSession :: LoadFile( const char * pchFile, int * piScanMode, bool * pbTimeSep, int iWid ) { 
	int i;
	
	//-- Open File
	FILE * h = fopen( pchFile, "rb" );
	if( h == NULL ) 
		return false;

	//-- Get Session header info -- version 1
	int iPanels, iGenFiles, iVersion;
	bool bReadError =  fread( &iVersion, sizeof(iVersion), 1, h ) != 1 
		|| ! FileLoadString( h, &m_cConfigName )
		|| fread( &m_iGenBufferSize, sizeof(m_iGenBufferSize), 1, h ) != 1 
		|| fread( &m_iGenRate, sizeof(m_iGenRate), 1, h ) != 1  
		|| fread( &iGenFiles, sizeof(iGenFiles), 1, h ) != 1  
		|| fread( &m_iAcqBufferSize, sizeof(m_iAcqBufferSize), 1, h ) != 1 
		|| fread( &m_iAcqRate, sizeof(m_iAcqRate), 1, h ) != 1 
		|| fread( &m_iAcqSampPerPixel, sizeof(m_iAcqSampPerPixel), 1, h ) != 1 
		|| ! FileLoadString( h, &m_cAcqFile ) 
		|| fread( &iPanels, sizeof(int), 1, h ) != 1 ;
		
	//-- version 2 - adds m_bRecordDigout
	BYTE btControls=0;
	bReadError |= ( iVersion>=2 && fread( &btControls, sizeof(BYTE), 1, h ) != 1 );
	m_bRecordDigout=(btControls&4)>0;
	
	//-- version 3 - adds time sep and scanmode
	*piScanMode=1;
	*pbTimeSep=false;
	bReadError |= ( iVersion>=3
					&& ( fread( piScanMode, sizeof(int), 1, h ) != 1 
						|| fread( pbTimeSep, sizeof(bool), 1, h ) != 1 )
					);

	//-- version 4 - adds gain table : flag, name, source, dest
	bReadError |= ( iVersion>=4 
					&& ( fread( &m_bGainTable, sizeof(bool), 1, h)!=1 
						|| ! FileLoadString( h, &m_cGainTable ) 
						|| fread( &m_iGainTableSource, sizeof(int), 1, h )!=1 
						|| fread( &m_iGainTableApply, sizeof(int), 1, h )!=1 
					 ) );

	//-- version 5 - adds feedback configuration : flag, name, source, dest
	bReadError |= ( iVersion>=5 
					&& ( fread( &m_bFeedback, sizeof(bool), 1, h)!=1 
						|| ! FileLoadString( h, &m_cFeedback ) 
						|| fread( &m_iFeedbackSource, sizeof(int), 1, h )!=1 
						|| fread( &m_iFeedbackApply, sizeof(int), 1, h )!=1 
					 ) );

	//-- error in header load : 
	if( bReadError ) { 
		fclose(h);
		return false;
		}		
		
	//-- Load list of gen info
	for( i=0; i<iGenFiles; ++i ) {
		if( ! FileLoadString( h, &(m_acGenFile[i]) ) 
				|| fread( &(m_abGenRepeat[i]), sizeof(m_abGenRepeat[i]), 1, h ) != 1 ) {
			fclose(h);
			return false;
			}
		}
		
	//-- Get Panels Array
	m_aPanels.DeleteAll();
	CSessionPanel * pPnl;

	for( i=0; i<iPanels; ++i ) {
		pPnl = new CSessionPanel(iWid, 0, 0);
		if( ! pPnl->LoadFromFile(h) ) {
			fclose(h);
			return false;
			}
		m_aPanels.AddItem( pPnl );
		}

	//-- Version 6 adds Digital Generation filename and repeat flag 		
	if( iVersion>=6 ) { 
		if( ! FileLoadString( h, &m_cDigGenFile ) 
				|| fread( &m_bDigGenRepeat, sizeof(m_bDigGenRepeat), 1, h ) != 1 ) {
			fclose(h);
			return false;
			}
		}

	//-- Close file
	fclose(h);
	return true;
	}

bool CSession :: SaveFile( const char * pchFile, int iScanMode, bool bTimeSep ) { 
	int i;
	//-- Open File
	FILE * h = fopen( pchFile, "wb" );
	if( h == NULL ) 
		return false;
	//-- Write Session header info 
	int iPanels = m_aPanels.Count();
	int iGenFiles = SESSION_MAXGENCHN;
	int iVersion=6;
	
	BYTE btControls = (m_bRecordDigout?4:0);
	if ( fwrite( &iVersion, sizeof(iVersion), 1, h ) != 1 
		|| ! FileSaveString( h, m_cConfigName.Buffer() )
		|| fwrite( &m_iGenBufferSize, sizeof(m_iGenBufferSize), 1, h ) != 1 
		|| fwrite( &m_iGenRate, sizeof(m_iGenRate), 1, h ) != 1  
		|| fwrite( &iGenFiles, sizeof(iGenFiles), 1, h ) != 1  
		|| fwrite( &m_iAcqBufferSize, sizeof(m_iAcqBufferSize), 1, h ) != 1 
		|| fwrite( &m_iAcqRate, sizeof(m_iAcqRate), 1, h ) != 1 
		|| fwrite( &m_iAcqSampPerPixel, sizeof(m_iAcqSampPerPixel), 1, h ) != 1 
		|| ! FileSaveString( h, m_cAcqFile.Buffer() ) 
		|| fwrite( &iPanels, sizeof(int), 1, h ) != 1
		//-- version 2 - add toggle controls
		|| fwrite( &btControls, sizeof(BYTE), 1, h ) != 1 
		//-- version 3 - add ScanMode and TimeSep
		|| fwrite( &iScanMode, sizeof(iScanMode), 1, h ) != 1 
		|| fwrite( &bTimeSep, sizeof(bTimeSep), 1, h ) != 1
		//-- version 4 - adds gain table : flag, name, source, dest
		|| fwrite( &m_bGainTable, sizeof(bool), 1, h)!=1 
		|| ! FileSaveString( h, m_cGainTable.ConstBuffer() ) 
		|| fwrite( &m_iGainTableSource, sizeof(int), 1, h )!=1 
		|| fwrite( &m_iGainTableApply, sizeof(int), 1, h )!=1 
		//-- version 5 - adds feedback configuration: flag, name, source, dest
		|| fwrite( &m_bFeedback, sizeof(bool), 1, h)!=1 
		|| ! FileSaveString( h, m_cFeedback.ConstBuffer() ) 
		|| fwrite( &m_iFeedbackSource, sizeof(int), 1, h )!=1 
		|| fwrite( &m_iFeedbackApply, sizeof(int), 1, h )!=1 
		) { 
		
		fclose(h);
		return false;
		}
	
	for( i=0; i<iGenFiles; ++i ) {
		if( ! FileSaveString( h, m_acGenFile[i].Buffer() ) 
				|| fwrite( &(m_abGenRepeat[i]), sizeof(m_abGenRepeat[i]), 1, h ) != 1 ) {
			fclose(h);
			return false;
			}
		}
	
	//-- Save Panels Arrays
	for( i=0; i<iPanels; ++i )
		if( ! ((CSessionPanel *)m_aPanels[i])->SaveToFile(h) ) 
			return false;

	//-- Version 6 adds Digital Generation filename and repeat flag 		
	if( ! FileSaveString( h, m_cDigGenFile.Buffer() ) 
			|| fwrite( &m_bDigGenRepeat, sizeof(m_bDigGenRepeat), 1, h ) != 1 ) {
		fclose(h);
		return false;
		}

	// add to recent menu
	g_oCfgs.m_oSettings.addSession( pchFile );

	//-- Close file
	fclose(h);
	return true;
	}

bool CSession :: LoadConfig() { 
	//-- find configuration
	int iCfg=g_oCfgs.FindConfig( m_cConfigName.Buffer() );
	if( -1 == iCfg ) {
		AddMessage( REPTYPE_ERROR, "Unlisted configuration specified in loaded session." );
		return false;
		}
	m_poCfg = g_oCfgs.Config(iCfg);
		
	//-- Resolve NI card pointer
	int iCard = NiCardFind( m_poCfg->m_cDeviceName.Buffer() );
	if( -1 == iCard ) 
		return false;
	m_poCard = NiCard( iCard );

	return true;
	}
	
CSessionPanel * CSession :: Panel( int iPanel ) {
	return (CSessionPanel *)m_aPanels[iPanel];
	}

//-- Build initial array of panels based on hardware configuration		
void CSession :: CreatePanels( int iHgt, int iWid, int iTop, int iLeft ) { 
	CChannelConfig * pCfg;
	CSessionPanel * pPnl;
	int iCh;

	//-- Analog Acquisition 	
	m_iAnaAcqPanels=0;
	for( iCh=0; iCh<CONFIG_AIROWS; ++iCh ) { 
		pCfg = &(m_poCfg->m_aoInChnCfg[iCh]);
		if( pCfg->m_iActive!=0 ) { 
			double dY = m_poCard->m_dInRange / pCfg->m_dGain;	// / pCfg->m_dScaling
			if ( pCfg->m_dScaling < 0 )
				dY = - dY;
			m_aPanels.AddItem(pPnl=new CSessionPanel( iWid, iLeft, iTop ) );
			pPnl->SetupAnaAcqPanel( pCfg->m_cChannelName.Buffer(), pCfg->m_cUnits.Buffer(), 
							iHgt, -1.0*dY, dY, -1.0 * dY, dY, true, iCh );
			pPnl->m_dScaling = pCfg->m_dScaling;
			iTop += pPnl->m_iHgt;
			++m_iAnaAcqPanels;
			}
		}
		
	//-- Digital Acquisition
	if( (m_iDigAcqLines=m_poCfg->DigAcqCount(&m_aiDigAcqMap)) > 0 ) { 
		m_aPanels.AddItem( pPnl=new CSessionPanel( iWid, iLeft, iTop ) );
		pPnl->SetupDigAcqPanel( m_poCfg->m_aiDigMode, 0 );
		iTop += pPnl->m_iHgt;
		}
		
	//-- Analog Generation
	m_iAnaGenPanels=0;
	int iGenPanel=0;
	for( iCh=0; iCh<CONFIG_AOROWS; ++iCh ) { 
		pCfg = &(m_poCfg->m_aoOutChnCfg[iCh]);
		if( pCfg->m_iActive!=0 ) { 
			m_aPanels.AddItem(pPnl=new CSessionPanel( iWid, iLeft, iTop ));
			pPnl->SetupAnaGenPanel( pCfg->m_cChannelName.Buffer(), pCfg->m_cUnits.Buffer(), true, iCh );
			m_aiGenChannelPanel[iCh]=m_aPanels.Count()-1;
			m_iAnaGenPanels++;
			iTop += pPnl->m_iHgt;
			}
		else 
			m_aiGenChannelPanel[iCh]=-1;
		}

	//-- Digital Generation
	if( (m_iDigGenLines=m_poCfg->DigGenProtocolCount(&m_aiDigGenMap)) > 0 ) { 
		m_aPanels.AddItem( pPnl=new CSessionPanel( iWid, iLeft, iTop ) );
		pPnl->SetupDigGenPanel( m_poCfg->m_aiDigMode, 0 );
		iTop += pPnl->m_iHgt;
		}
		
	}
	
void CSession :: SetupPanels( ) { 
	CChannelConfig * pCfg;
	CSessionPanel * pPnl;

	m_iAnaAcqPanels=0;
	m_iAnaGenPanels=0;
	m_iDigAcqLines=0;
	m_iDigGenLines=0;
	for( int iPnl=0; iPnl<m_aPanels.Count(); ++iPnl ) { 
		pPnl = Panel(iPnl);

		if( pPnl->m_iChannelType==CT_AI ) { 
			//-- Analog Acquisition
			pCfg = &(m_poCfg->m_aoInChnCfg[pPnl->m_iChannelIndex]);
			double dY = m_poCard->m_dInRange  / pCfg->m_dGain; // / pCfg->m_dScaling
			if ( pCfg->m_dScaling < 0 ) {
				dY = - dY;
				if ( pPnl->m_dMinDisp < pPnl->m_dMaxDisp ) {
					double swapp = pPnl->m_dMaxDisp;
					pPnl->m_dMaxDisp = pPnl->m_dMinDisp;
					pPnl->m_dMinDisp = swapp;
				}
			}
			pPnl->SetupAnaAcqPanel( pCfg->m_cChannelName.Buffer(), pCfg->m_cUnits.Buffer(), 
						pPnl->m_iHgt, -1.0*dY, dY, 
						pPnl->m_dMinDisp, pPnl->m_dMaxDisp, pCfg->m_iActive!=0, pPnl->m_iChannelIndex );
			pPnl->m_dScaling = pCfg->m_dScaling;
			++m_iAnaAcqPanels;
			}

		else if( pPnl->m_iChannelType==CT_AO ) { 
			//-- Analog Generation
			pCfg = &(m_poCfg->m_aoOutChnCfg[pPnl->m_iChannelIndex]);
			pPnl->SetupAnaGenPanel( pCfg->m_cChannelName.Buffer(), pCfg->m_cUnits.Buffer(), pCfg->m_iActive!=0, pPnl->m_iChannelIndex );
			m_aiGenChannelPanel[pPnl->m_iChannelIndex]=iPnl;
			m_iAnaGenPanels++;
			}
			
		else if( pPnl->m_iChannelType==CT_DI && m_iDigAcqLines==0 ) { 
			//-- Digital Acquisition
			m_iDigAcqLines=m_poCfg->DigAcqCount(&m_aiDigAcqMap);
			pPnl->SetupDigAcqPanel( m_poCfg->m_aiDigMode, 0 );
			}

		else if( pPnl->m_iChannelType==CT_DO && m_iDigGenLines==0 ) { 
			//-- Digital Generation
			m_iDigGenLines=m_poCfg->DigGenProtocolCount(&m_aiDigGenMap);
			pPnl->SetupDigGenPanel( m_poCfg->m_aiDigMode, 0 );
			}
		}
	}

//-- Calculate the individual display Y ranges from the main disp rectangle
void CSession :: SetPanelPositions( int iTop ) { 
	int * aiPage = new int [m_iPages];
	for( int iPage=0; iPage<m_iPages; ++iPage ) 
		aiPage[iPage]=iTop;
		
	for( int iPanel=0; iPanel<m_aPanels.Count(); ++iPanel ){ 
		CSessionPanel * pPnl = Panel(iPanel);
		if( pPnl->m_iPage>=0 && pPnl->m_iPage<m_iPages ) {
			pPnl->m_iTop=aiPage[pPnl->m_iPage];
			aiPage[pPnl->m_iPage]+=pPnl->m_iHgt;
			//pPnl->m_iBtm=aiPage[pPnl->m_iPage];
			}
		}
	}

bool CSession :: CreateFeedback() { 
	if( m_pFeedback != NULL ) {
		delete m_pFeedback;
		m_pFeedback=NULL;
		}
		
	//-- setup feedback application
	if( m_bFeedback ) {
		//-- check if specified feedback configuration exists 
		if( (m_pFeedback = g_oCfgs.FeedbackConfig(m_cFeedback.ConstBuffer()) ) != NULL ) 
			m_pFeedback = m_pFeedback->NewCopy();
		}
	return true;
	}
	
//-- if not yet created, create the .qdf file.
// Allocate space for digital and analog, even if only one is in use.
void CSession :: AcqOpenFile() { 
	int iLine;
	if( ! m_oQDF.Opened() ) {
		m_oQDF.CreateQdf( m_cAcqFile.Buffer(), 1.0/double(m_iAcqRate), 
			m_poAnaAcqTask->m_iAcqChannels + m_poAnaAcqTask->m_iDigAcqLines );
		
		//-- add qsf file initializer info 
		m_oQDF.GoRoot();
		m_oQDF.AddNode( "SessionFile", true, QTP_EMPTY, 0, 0 );
		m_oQDF.AddNode( "DataChannels", true, QTP_EMPTY, 0, 0 );

		for( int iPnl=0; iPnl<m_aPanels.Count(); ++iPnl ) {
			CSessionPanel * pPnl = Panel(iPnl);
			
			//-- Analog Acquisition
			if( pPnl->m_iChannelType==CT_AI  ) { // && pPnl->m_bActive  remove bActive ? 
				CChannelConfig * pChCfg = &(m_poCfg->m_aoInChnCfg[pPnl->m_iChannelIndex]);
				m_oQDF.AddChannel( pPnl->m_cName.Buffer(), pPnl->m_cUnits.Buffer(), pPnl->m_dMax/10.0 * pChCfg->m_dScaling, pChCfg->m_dGain);
				}
				
			//-- Digital Acquisition
			if( pPnl->m_iChannelType==CT_DI  ) 
				for( iLine=0; iLine<m_iDigAcqLines; ++iLine ) 
					m_oQDF.AddChannel( m_poCfg->m_acDigName[m_aiDigAcqMap[iLine]].Buffer(), "On/Off", 1.0, 1.0 );
					
			//-- Digital Generation
			//if( pPnl->m_iChannelType==CT_DO && m_bRecordDigout ) 
			//	for( iLine=0; iLine<m_iDigGenLines; ++iLine ) 
			//		m_oQDF.AddChannel( m_poCfg->m_acDigName[m_aiDigGenMap[iLine]].Buffer(), "Set", 1.0, 1.0 );
			}
			
		//-- add hardware configuration as a comment node.
		int i;
		char buf[4096];
		sprintf( buf, "Hardware configuration (name, units, gain, scaling ):\n");
		for( i=0; i<CONFIG_AIROWS; ++i ) 
			if( m_poCfg->m_aoInChnCfg[i].m_iActive ) {
				sprintf( buf+strlen(buf), "AI%d :  '%s'  '%s'  %f  %f \n", 
					i, 
					m_poCfg->m_aoInChnCfg[i].m_cChannelName.Buffer(), 
					m_poCfg->m_aoInChnCfg[i].m_cUnits.Buffer(), 
					m_poCfg->m_aoInChnCfg[i].m_dGain, 
					m_poCfg->m_aoInChnCfg[i].m_dScaling
					);				
				}

		for( i=0; i<CONFIG_AOROWS; ++i ) 
			if( m_poCfg->m_aoOutChnCfg[i].m_iActive ) 
				sprintf( buf+strlen(buf), "AO%d :  '%s'  '%s'  %f  %f \n", 
					i,
					m_poCfg->m_aoOutChnCfg[i].m_cChannelName.Buffer(), 
					m_poCfg->m_aoOutChnCfg[i].m_cUnits.Buffer(), 
					m_poCfg->m_aoOutChnCfg[i].m_dGain, 
					m_poCfg->m_aoOutChnCfg[i].m_dScaling
					);				

		if( m_iDigAcqLines>0 || m_iDigGenLines>0 ) { 
			sprintf( buf+strlen(buf), "Digital " );
			for( iLine=0; iLine<CONFIG_DOROWS; ++iLine ) 
				if( m_poCfg->m_aiDigMode[iLine] > DOMODE_DISABLED ) 
					sprintf( buf+strlen(buf), "[%d]=%s ", iLine, g_apchDigMode[m_poCfg->m_aiDigMode[iLine]] );
			sprintf( buf+strlen(buf), "\n" );
			}

		if( m_bGainTable && m_pGainTable!=NULL ) { 
			sprintf( buf+strlen(buf), "GainTable:%s from %d to %d {", m_pGainTable->m_cTableName.ConstBuffer(), m_iGainTableSource, m_iGainTableApply );
			for( i=0; i<m_pGainTable->m_iEntries; ++i ) 
				sprintf( buf+strlen(buf), "%f:%f%s", 
						m_pGainTable->m_adSignal[i], m_pGainTable->m_adMultiplier[i], 
						i==m_pGainTable->m_iEntries-1 ? "}\n" : ", " );
			}
		m_oQDF.WriteComment( buf );

		//-- restore ptr to data segments node			
		m_oQDF.GoRoot();
		m_oQDF.GoChild("Segments");
		}
	}

bool CSession :: Go( int iQueueSize, bool bAcq, bool bGen, bool bRecord ) { 
	//-- setup acquisition gain table application
	if( m_bGainTable ) {
		//-- check if specified gain table exists 
		if( (m_pGainTable = g_oCfgs.GainTable(m_cGainTable.ConstBuffer()) ) != NULL ) 
			m_pGainTable = new CGainTable(*m_pGainTable);	//-- make a copy of the gain table.
		if( m_pGainTable != NULL ) {
			m_pGainTable->m_iCurrent=-1;
			}
		}
	else 
		m_pGainTable=NULL;

	//-- clear prior digital outputs array ( should already be empty though )					
	m_aDigout.DeleteAll();
							
	//-- Setup task - this new is deleted by the poller 
	m_poAnaAcqTask = new CTask( m_poCard, 
				bAcq && m_iAnaAcqPanels>0, m_iAcqRate, m_iAcqBufferSize, 
				bGen && m_iAnaGenPanels>0, m_iGenRate, m_iGenBufferSize, 
				m_bFeedback, 
				bAcq && m_iDigAcqLines>0, 
				bGen && m_iDigGenLines>0 );
		
	//-- setup output channels
	int i;
	if( bGen ) { 
		m_iGenChannel=-1;
		//-- Analog output channels
		for( i=0; i<CONFIG_AOROWS; ++i )
			if( m_poCfg->m_aoOutChnCfg[i].m_iActive!=0 ) {
				if( ! m_poAnaAcqTask->AddOutputChannel( i ) || ! m_aProtocol[i].GenPrepare(m_iGenRate) ) {
					AddMessage( REPTYPE_ERROR, "Error creating output channel" );
					return false;
					}
				if( m_bFeedback && i==m_iFeedbackApply ) 
					m_pFeedback->m_iOutputIndex=m_poAnaAcqTask->m_iGenChannels-1;
				}

		//-- Digital output channel
		if( m_iDigGenLines > 0 )
			if( ! m_poAnaAcqTask->AddDigitalOutputChannels( 0,  m_iDigGenLines, m_aiDigGenMap ) 
				|| ! m_oDigProtocol.GenPrepare(m_iGenRate) ) {
				AddMessage( REPTYPE_ERROR, "Error creating digital output channel" );
				return false;
				}

		m_iAnaGenPos=m_iDigGenPos=0;
		}

	//-- setup input channels
	int iMode = m_poCfg->m_iAcqMeasMode==2 
		? DAQmx_Val_Diff 
		: ( m_poCfg->m_iAcqMeasMode==1 ? DAQmx_Val_NRSE : DAQmx_Val_RSE );
	
	m_iGainSourceIndex=-1;
	m_iGainApplyIndex=-1;
	if( bAcq ) { 
		for( i=0; i<CONFIG_AIROWS; ++i )
			if( m_poCfg->m_aoInChnCfg[i].m_iActive!=0 ) {
				if( ! m_poAnaAcqTask->AddInputChannel( i, m_poCfg->m_aoInChnCfg[i].m_dGain, iMode ) ) { 
					AddMessage( REPTYPE_ERROR, "Error creating input channel" );
					return false;
					}
				if( m_bGainTable && i==m_iGainTableSource ) 
					m_iGainSourceIndex=m_poAnaAcqTask->m_iAcqChannels-1;
				if( m_bGainTable && i==m_iGainTableApply ) 
					m_iGainApplyIndex=m_poAnaAcqTask->m_iAcqChannels-1;
				if( m_bFeedback && i==m_iFeedbackSource ) 
					m_pFeedback->m_iInputIndex=m_poAnaAcqTask->m_iAcqChannels-1;
				}
				
		//-- Digital acquisition channel
		if( m_iDigAcqLines > 0 )
			if( ! m_poAnaAcqTask->AddDigitalInputChannels( 0,  m_iDigAcqLines, m_aiDigAcqMap ) ) {
				AddMessage( REPTYPE_ERROR, "Error creating digital input channel" );
				return false;
				}

		//-- setup short term queue
		m_iQueueSize=iQueueSize;
		m_adQueue = new double[m_poAnaAcqTask->m_iAcqChannels * m_iQueueSize];
		m_aiDigQueue = new int[m_iQueueSize];
		m_iQueuePos = 0;
		
		//-- Add a new node in the output file to hold data 	
		AcqOpenFile();	
		if( bRecord ) 
			StartRecording();
		}

	//-- check for proper gain table setup		
	if( m_bGainTable && ( m_iGainSourceIndex==-1 || m_iGainApplyIndex==-1 ) ) {
		AddMessage( REPTYPE_ERROR, "Gain table specifies a channel which is not being recorded." );
		m_pGainTable=NULL;
		}
			
	//-- check for proper feedback configuration setup		
	if( m_bFeedback && ( m_pFeedback->m_iInputIndex==-1 || m_pFeedback->m_iOutputIndex==-1 ) ) {
		AddMessage( REPTYPE_ERROR, "Feedback specifies a channel which is not active." );
		m_pFeedback=NULL;
		}
		
	//-- Fire up the acquisition board 
	m_pAcqInit = CreateInit( AcqDispatch, GenDispatch, TermAcqDispatch, this );
	m_poAnaAcqTask->StartTasks( m_pAcqInit, false );
	
	return true;
	} 

//-- Callback to process data from acquired buffer to session 
bool CSession :: AcqData( int iSamples, void * pAnaData, void * pDigData ) { 
	if( ! m_oQDF.Opened() ) 
		return false;

	double aError[16]={ DATA_GLITCH, DATA_GLITCH, DATA_GLITCH, DATA_GLITCH,
						DATA_GLITCH, DATA_GLITCH, DATA_GLITCH, DATA_GLITCH,
						DATA_GLITCH, DATA_GLITCH, DATA_GLITCH, DATA_GLITCH,
						DATA_GLITCH, DATA_GLITCH, DATA_GLITCH, DATA_GLITCH };

	//-- Ignore data if > 60% of buffer is full.  Add a 'Glitch' record to the data.
	// This is an attempt to keep the acq running at the expense of some data when it 
	// is getting behind.
	//  TODO - allow > 60% ? 
	//  TODO - Consider implications of lost data on recorded digital output ? 
	//  TODO - add a note or empty section to data ? 	
	if( iSamples > MulDiv( m_iAcqBufferSize, 6, 10 ) ) {
		iSamples=1;
		pAnaData=aError;
		}

	int iChans=m_poAnaAcqTask->m_iAcqChannels;

	//-- apply gain table multiplier 
	if( m_pGainTable!=NULL ) { 
		double * pdSource = ((double *)pAnaData)+m_iGainSourceIndex;
		double * pdApply = ((double *)pAnaData)+m_iGainApplyIndex;
		for( int i=0; i<iSamples; ++i ) {
			*pdApply *= m_pGainTable->Multiplier( *pdSource );
			pdApply+=iChans;
			pdSource+=iChans;
			}
		}

	//-- copy data to qdf file	
	if( m_bRecording ) {
		m_iAcqPos += iSamples;
		m_oQDF.AddData( (double *) pAnaData, iChans, (int *) pDigData, m_iDigAcqLines, 
						m_aiDigAcqMap, iSamples, false );	
		}

	//todo - if samples count is approaching card buffersize, skip the display 
	//----- write to the queue
	//-- limit Write to scope queue to <= queue size
	if( iSamples>m_iQueueSize ) {
		if( pAnaData!=NULL ) 
			pAnaData=((BYTE *)pAnaData)+(iSamples-m_iQueueSize)*iChans*sizeof(double);
		if( pDigData!=NULL ) 
			pDigData=((BYTE *)pDigData)+(iSamples-m_iQueueSize)*sizeof(int);
		iSamples=m_iQueueSize;
		}
	
	//-- if wrap will occur, write the first portion to the end of the que
	if( m_iQueuePos+iSamples > m_iQueueSize ) { 
		int iCopy = (m_iQueueSize-m_iQueuePos);
		if( pAnaData!=NULL ) {
			memcpy( &(m_adQueue[m_iQueuePos*iChans]), pAnaData, iCopy*sizeof(double)*iChans );
			pAnaData=((BYTE *)pAnaData)+iCopy*sizeof(double)*iChans;
			}
		if( pDigData!=NULL ) {
			memcpy( &(m_aiDigQueue[m_iQueuePos]), pDigData, iCopy*sizeof(int) );
			pDigData=((BYTE *)pDigData)+iCopy*sizeof(int);
			}
		m_iQueuePos=0;
		iSamples -= iCopy;
		}
	//-- write remaining samples to the queue - this is the general write.  
	if( iSamples > 0 ) { 
		if( pAnaData!=NULL )
			memcpy( &(m_adQueue[m_iQueuePos*iChans]), pAnaData, iSamples*iChans*sizeof(double) );
		if( pDigData!=NULL )
			memcpy( &(m_aiDigQueue[m_iQueuePos]), pDigData, iSamples*sizeof(int) );
		m_iQueuePos += iSamples;
		}
				
	//-- Cause View window update
	PostMessage( m_hView, WM_DATA_UPDATE, m_bRecording?1:0, 0 );
	return true;
	} 

//-- Start Acq / Gen of data with save
bool CSession :: Record( int iQueueSize ) { 
	Go( iQueueSize, true, false, true );
	return true;
	} 

//-- Start Acq of data without save
bool CSession :: Play( int iQueueSize ) { 
	Go( iQueueSize, true, false, false );
	return true;
	} 

//-- Start Acq / Gen of data 
bool CSession :: Synch( int iQueueSize ) { 
	return Go( iQueueSize, true, true, true );
	} 

//-- Stop Acq / Gen of data 
bool CSession :: Stop() {
	m_pAcqInit->m_bCancelled=true;
	return true;
	}

bool CSession :: Generate( int iCh, bool bContinuous ) { 
	bool bAnaGen = ( iCh==-1 && m_iAnaGenPanels>0 ) 
				|| ( iCh>=0 && iCh<CONFIG_AOROWS && m_poCfg->m_aoOutChnCfg[iCh].m_iActive );
	bool bDigGen = ( ( iCh==-1 || iCh==DIGGENCHANNELID ) && m_iDigGenLines>0 ) ;
	if( ! bAnaGen && ! bDigGen ) 
		return false;
	int iGenRate = bAnaGen ? m_iGenRate : 100000;
	m_iGenChannel=iCh;

	//-- Setup task - this new is deleted by the poller 
	int iLen=m_iGenBufferSize;
	if( iCh==-1 )  
		bContinuous=true;
	else if( iCh==DIGGENCHANNELID ) { 
		if( ! bContinuous ) 
			iLen=m_oDigProtocol.SampleCount(iGenRate);
		}
	else if( iCh>=0 && iCh<CONFIG_AOROWS ) 
		iLen = m_aProtocol[iCh].TotalSamples(iGenRate);
	
	m_poAnaGenTask = new CTask( m_poCard,  false, 0, 0,  bAnaGen, iGenRate, iLen, false, false, bDigGen );
			 // Note - no ffedback with acq stopped or non synchronized.  TODO - consider non-synchronized...
	m_poAnaGenTask->m_b100khz=( ! bAnaGen );
	 
	//-- setup Analog output channels
	if( bAnaGen ) { 
		if( m_bFeedback && m_pFeedback!=NULL )
			m_pFeedback->m_iOutputIndex=-1;
		
		for( int i=0; i<CONFIG_AOROWS; ++i ) 
			if( (iCh==-1 || iCh==i) && (m_poCfg->m_aoOutChnCfg[i].m_iActive!=0) ) {
				if( ! m_poAnaGenTask->AddOutputChannel( i ) 
						|| ! m_aProtocol[i].GenPrepare(iGenRate) ) {
					AddMessage( REPTYPE_ERROR, "Error creating output channel" );
					return false;
					}
				if( m_bFeedback && i==m_iFeedbackApply ) 
					m_pFeedback->m_iOutputIndex=m_poAnaAcqTask->m_iGenChannels-1;
				}
		}
					
	//-- setup Digital output channels
	if( bDigGen ) { 
		if( ! m_poAnaGenTask->AddDigitalOutputChannels( 0,  m_iDigGenLines, m_aiDigGenMap ) ) {
			AddMessage( REPTYPE_ERROR, "Error creating digital output channel" );
			return false;
			}
		m_oDigProtocol.GenPrepare( iGenRate );
		}
		
	//-- Rewind 
	m_iAnaGenPos=m_iDigGenPos=0;

	//-- Fire up the acquisition board 
	m_pGenInit = CreateInit( NULL, GenDispatch, TermGenDispatch, this );
	m_poAnaGenTask->StartTasks( m_pGenInit, ! bContinuous );
	
	return true;
	}
	
bool CSession :: GenStop( ) { 
	if( m_pGenInit!=NULL ) 
		m_pGenInit->m_bCancelled=true;
	return true;
	}
	
//-------------------------------------------------------------------------------
//-- Called from poller when no more data will be sent or requested
bool CSession :: AcqCompleted( ) { 
	if( m_adQueue != NULL ) {
		delete [] m_adQueue;
		m_adQueue=NULL;
		}

	if( m_aiDigQueue != NULL ) {
		delete [] m_aiDigQueue;
		m_aiDigQueue=NULL;
		}
		
	if( m_bRecording ) 
		EndRecording();
	
	if( m_pGainTable != NULL ) {
		delete m_pGainTable;
		m_pGainTable=NULL;
		}

	//-- notify view that polling is complete
	PostMessage( m_hView, WM_ACQPOLLSTOP, 0, 0 );

	return true;
	}

bool CSession :: GenCompleted( ) { 
	m_iGenChannel=0;
	m_poAnaGenTask=NULL;		// delete is performed by the poller

	// notify view that polling is complete
	PostMessage( m_hView, WM_GENPOLLSTOP, 0, 0 );

	return true;
	}

//-- Callback to generate data from current session into buffer
bool CSession :: GenData( int iSamples, void * pData, void * pDigData, int iSetPos ) { 
	if( iSetPos != -1 ) 
		m_iAnaGenPos=m_iDigGenPos=iSetPos;

	//-- Analog
	if( pData==NULL ) {
		}
	else if( m_poAnaGenTask == NULL || m_iGenChannel==-1 ) {	// generate all channels
		int iGenCh=0;				// index of active Generation channels - for buffer loading
		for( int iCh=0; iCh<CONFIG_AOROWS; ++iCh ) {
			if( m_poCfg->m_aoOutChnCfg[iCh].m_iActive ) {
				m_aProtocol[iCh].GenerateData( &(((double *)pData)[iGenCh*iSamples]), iSamples, 
						m_abGenRepeat[iCh], m_poCfg->m_aoOutChnCfg[iCh].m_dScaling, 
						-1.0 * m_poCard->m_dOutRange, m_poCard->m_dOutRange);
				++iGenCh;
				}
			}
		m_iAnaGenPos+=iSamples;
		}
	else if( m_iGenChannel>=0 && m_iGenChannel<CONFIG_AOROWS ) {	//-- generate 1 channel
		m_aProtocol[m_iGenChannel].GenerateData( (double *)pData, iSamples, m_abGenRepeat[m_iGenChannel], 
				m_poCfg->m_aoOutChnCfg[m_iGenChannel].m_dScaling, 
				-1.0 * m_poCard->m_dOutRange, m_poCard->m_dOutRange );
		m_iAnaGenPos+=iSamples;
		}

	//-- Digital 
	if( pDigData!=NULL && ( m_iGenChannel==DIGGENCHANNELID || ( m_iGenChannel==-1 && m_iDigGenLines!=0 ) ) ) {
		m_oDigProtocol.GenerateData( (unsigned int *)pDigData, iSamples, m_bDigGenRepeat );
		m_iDigGenPos+=iSamples;
		}
		
	return true;
	}

bool CSession :: LoadProtocols() {
	int iActive=0;
	for( int i=0; i<CONFIG_AOROWS; ++i )
		if( m_poCfg->m_aoOutChnCfg[i].m_iActive!=0 ) { 
			int iCmaPos1 = m_acGenFile[i].Find(',');
			int iCmaPos2 = m_acGenFile[i].Find(',', iCmaPos1+1);
			if( iCmaPos1>=5 && iCmaPos2>=7 && m_acGenFile[i].Find(".qdp")>0 ) { 
				CDigitalProtocol o;
				if( ! o.LoadFile(m_acGenFile[i].Left(iCmaPos1).Buffer() ) )
					AddMessage( REPTYPE_ERROR, "Error loading protocol file from qdp" );
				else { 
					int iLine = atoi_l( m_acGenFile[i].ConstBuffer()+iCmaPos1+1, iCmaPos2-iCmaPos1-1 );
					double dLevel = atof_l( m_acGenFile[i].ConstBuffer()+iCmaPos2+1, m_acGenFile[i].Length()-iCmaPos2-1 );
					ProtocolDigitalToAnalog( &o, iLine, dLevel, &(m_aProtocol[i]) );
					}
				}
			else { 
				if( ! m_aProtocol[i].LoadFile( m_acGenFile[i].Buffer() ) ) 
					AddMessage( REPTYPE_ERROR, "Error loading protocol file" );
				}
			iActive++;		
			}

	bool bDigLoad=false;			
	if( m_poCfg->HasProtocolDigGen() )
		if( ! ( bDigLoad=m_oDigProtocol.LoadFile( m_cDigGenFile.Buffer() ) ) )
			AddMessage( REPTYPE_ERROR, "Error loading protocol file" );
			
	//---- Save provided file names to registry 
	if( iActive > 0 ) { 
		char buf[_MAX_PATH+1], bufName[16];
		buf[_MAX_PATH]=0;
		HKEY h;
		RegCreateKeyEx( HKEY_CURRENT_USER, "Software\\QubIO", 0, NULL, 0, KEY_ALL_ACCESS, NULL, &h, NULL );

		for( int iCh=0; iCh<4; ++iCh ) { 
			if( m_acGenFile[iCh].TrimLength()>0 ) { 
				sprintf( bufName, "GenFile%d", iCh );
				buf[0]=m_abGenRepeat[iCh]?'Y':'N';
				strcpy( buf+1, m_acGenFile[iCh].Buffer() );
				RegSetValueEx( h, bufName, 0, REG_SZ, (BYTE *)buf, strlen(buf)+1 );	
				}
			}
		
		if( bDigLoad && m_cDigGenFile.TrimLength()>0 ) {
			buf[0]=m_bDigGenRepeat?'Y':'N';
			strcpy( buf+1, m_cDigGenFile.Buffer() );
			RegSetValueEx( h, "DigGenFile", 0, REG_SZ, (BYTE *)buf, strlen(buf)+1 );	
			}

		RegCloseKey( h );
		}
	
	return true;
	}

CDigoutEvent * CSession :: DigoutEvent( int iEvent ) { 
	return (CDigoutEvent *)m_aDigout[iEvent];
	}

void CSession :: RecordDigout( int iDigout, int iStatus ) { 
	if( m_bRecording && m_bRecordDigout ) 
		m_aDigout.AddItem( new CDigoutEvent( m_iAcqPos, iDigout, iStatus ) );
	}
	
void CSession :: StartRecording() { 
	m_bRecording=true;
	m_oQDF.AddDataNode( m_poAnaAcqTask->m_iAcqChannels + m_poAnaAcqTask->m_iDigAcqLines );
	}
	
void CSession :: EndRecording() { 
	//-- Flush data node	
	if( m_bRecording ) { 
		m_oQDF.AddData( NULL, m_poAnaAcqTask->m_iAcqChannels, NULL, m_iDigAcqLines, m_aiDigAcqMap, 0, true );	
		m_bRecording=false;
		}
	
	//-- Write digout node ( assumes current node is "Channels" in "Segment" in "Segments".
	if( m_bRecordDigout && m_aDigout.Count()>0 ) { 
		int iEvent, iEvents=m_aDigout.Count();
		
		//-- build array of digital output events
		unsigned int * pBuf = new unsigned int[	2*iEvents+2 ];
		pBuf[0]=m_iDigoutStatus;					// mask of which lines are active.
		
		//-- mask of which lines are auto-reset
		pBuf[1]=0;
		for( int i=0; i<32 && i<CONFIG_DOROWS; ++i ) 
			if( m_poCfg->m_aiDigMode[i]==DOMODE_AUTO ) 
				pBuf[1] += (1<<i);	
		
		//-- copy events{time,line} to buffer
		for( iEvent=0; iEvent<iEvents; ++iEvent ) { 
			CDigoutEvent * pEvent = DigoutEvent(iEvent);
			pBuf[iEvent*2+2]=pEvent->m_iAcqPos;
			pBuf[iEvent*2+3]=pEvent->m_iLine+(pEvent->m_iStatus<<16);
			}
		
		//-- add digout node to qdf file
		m_oQDF.AddDigoutNode( 2*iEvents+2, pBuf );
		
		//-- cleanup 
		delete [] pBuf;
		}
		
	fflush(m_oQDF.m_hFile);
	
	m_aDigout.DeleteAll();
	}
	
//-- Draw the session background	
void CSession :: DrawBgr( HDC hdc, int iPanel )  { 
	int iPanels = m_aPanels.Count();
	for( int iPnl=0; iPnl<iPanels; ++iPnl )
		if( Panel(iPnl)->m_iPage==m_iPage  )
			Panel(iPnl)->DrawExtra( hdc, m_poCfg );
	}

//-- Draw the data traces
void CSession :: DrawFgr( HDC hdc, int iScanMode )  { 
	int iPanels = m_aPanels.Count();
	for( int iPanel=0; iPanel<iPanels; ++iPanel ) { 
		CSessionPanel * pPnl = (CSessionPanel *)(m_aPanels[iPanel]);
		if( ! pPnl->m_bIconized ) 
			if( iScanMode==1 ) { 
				BitBlt( hdc, 
						pPnl->m_iLft, 
						pPnl->m_iTop+1, 
						pPnl->m_iWid-pPnl->m_iBitmapOffset, 
						pPnl->m_iHgt-1, 
						pPnl->m_hdcBitmap, pPnl->m_iBitmapOffset, 0, SRCCOPY );
				BitBlt( hdc,
						pPnl->m_iLft+pPnl->m_iWid-pPnl->m_iBitmapOffset, 
						pPnl->m_iTop+1, 
						pPnl->m_iBitmapOffset, 
						pPnl->m_iHgt-1, 
						pPnl->m_hdcBitmap, 0, 0, SRCCOPY );
				}
			else
				BitBlt( hdc, 
						pPnl->m_iLft, 
						pPnl->m_iTop+1, 
						pPnl->m_iWid, 
						pPnl->m_iHgt-1, 
						pPnl->m_hdcBitmap, 0, 0, SRCCOPY );
		}
	}



